package com.sillycycle.bagleyd.threed;

/*
 * @(#)ThreedMath.java
 *
 * Copyright 2005 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Point;

/**
 * The <code>ThreedMath</code> class handles the graphical
 * mathematics for the <code>ThreedFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/threed.html
 */

public class ThreedMath {
	static final int TOP = 0;
	static final int RIGHT = 1;
	static final int BOTTOM = 2;
	static final int LEFT = 3;
	//double[] sine = new double[ThreedInterface.RT_ANGLE + 1];
	int pixX, pixY, pixNo;

	/* Black Art of 3D Game Programming Andre LaMothe */

	static boolean clipPoint(Point a, Point b, Point minClip, Point maxClip) {
		Point intersection = new Point(0, 0);
		boolean[] edges = new boolean[4];
		boolean success = false;
		double dx, dy;

		edges[0] = edges[1] = edges[2] = edges[3] = false;
		/* compute deltas */
		dx = (double) b.x - a.x;
		dy = (double) b.y - a.y;
		/* compute what boundary lines need to be clipped against */
		/* vertical intersections */
		if (b.x > maxClip.x) {
			edges[RIGHT] = true;
			if (dx == 0.0) /* invalidate intersection */
				intersection.y = -1;
			else
				intersection.y = (int) (0.5 + (dy / dx) *
					(maxClip.x - a.x) + a.y);
		} else if (b.x < minClip.x) {
			edges[LEFT] = true;
			if (dx == 0.0) /* invalidate intersection */
				intersection.y = -1;
			else
				intersection.y = (int) (0.5 + (dy / dx) *
					(minClip.x - a.x) + a.y);
		}
		/* horizontal intersections */
		if (b.y > maxClip.y) {
			edges[BOTTOM] = true;
			if (dy == 0.0) /* invalidate intersection */
				intersection.x = -1;
			else
				intersection.x = (int) (0.5 + (dx / dy) *
					(maxClip.y - a.y) + a.x);
		} else if (b.y < minClip.y) {
			edges[TOP] = true;
			if (dy == 0.0) /* invalidate intersection */
				intersection.x = -1;
			else
				intersection.x = (int) (0.5 + (dx / dy) *
					(minClip.y - a.y) + a.x);
		}
		/* now we know where the line passed through */
		/* compute which edge is the proper intersection */
		if (edges[RIGHT] && (intersection.y >= minClip.y &&
				intersection.y <= maxClip.y)) {
			b.x = maxClip.x;
			b.y = intersection.y;
			success = true;
		} else if (edges[LEFT] && (intersection.y >= minClip.y &&
				intersection.y <= maxClip.y)) {
			b.x = minClip.x;
			b.y = intersection.y;
			success = true;
		}
		if (edges[BOTTOM] && (intersection.x >= minClip.x &&
				intersection.x <= maxClip.x)) {
			b.y = maxClip.y;
			b.x = intersection.x;
			success = true;
		} else if (edges[TOP] && (intersection.x >= minClip.x &&
				intersection.x <= maxClip.x)) {
			b.y = minClip.y;
			b.x = intersection.x;
			success = true;
		}
		return success;
	}

	static boolean clipLine(Point a, Point b, Point minClip, Point maxClip) {
		boolean aSuccess = false, bSuccess = false;
		boolean aVisible = false, bVisible = false;

		/* test if line completely visible */
		if ((a.x >= minClip.x && a.x <= maxClip.x) &&
				(a.y >= minClip.y && a.y <= maxClip.y))
			aVisible = true;
		if ((b.x >= minClip.x && b.x <= maxClip.x) &&
				(b.y >= minClip.y && b.y <= maxClip.y))
			bVisible = true;
		if (aVisible && bVisible)
			return true;

		if (!aVisible && !bVisible &&
				(((a.x < minClip.x && b.x < minClip.x) ||
				(a.x >= minClip.x && b.x > maxClip.x)) ||
				((a.y >= minClip.y && b.y < minClip.y) ||
				(a.y >= minClip.y && b.y > maxClip.y))))
			/* test if line completely invisible */
			return false;
		/* take care of case where either endpoint is in clipping region */
		if (!bVisible)
			bSuccess = clipPoint(a, b, minClip, maxClip);
		/* test second endpoint */
		if (!aVisible)
			aSuccess = clipPoint(b, a, minClip, maxClip);
		return (aSuccess || bSuccess);
	}

	/* Arash Partow <arashp AT hotmail.com>
	 * Homepage: http://www.partow.net
	 * Translated to C from Pascal
	 */
	static int orientation(Point a, Point b, Point p) {
		 /* Linear determinant of the 3 points */
		int orient = (b.x - a.x) * (p.y - a.y)
			 - (p.x - a.x) * (b.y - a.y);
		if (orient > 0) {
			return 1; /* Orientation is to the right-hand side */
		} else if (orient < 0) {
			return -1; /* Orientation is to the left-hand side */
		} else {
			return 0; /* Orientation is neutral if result is 0 */
		}
	}

	static boolean pointInTriangle(Point pt, Point a, Point b, Point c) {
		int or0 = orientation(a, b, pt);
		int or1 = orientation(b, c, pt);
		int or2 = orientation(c, a, pt);

		return (((or0 == or1) && (or1 == or2)) ||
			((or0 == 0) && (or1 == or2)) ||
			((or1 == 0) && (or0 == or2)) ||
			((or2 == 0) && (or0 == or1)));
	}

	static boolean pointInPolygon(Point pt, Point[] poly, int n) {
		int tri;
		Point a, b, c;

		a = poly[0];
		b = poly[1];
		for (tri = 2; tri < n; tri++) {
			c = poly[tri];
			if (pointInTriangle(pt, a, b, c)) {
				return true;
			}
			b = c;
		}
		return false;
	}

	boolean bresenhamLine(int startX, int startY, int endX, int endY) {
		int pix = 0;
		int ex = endX - startX;
		int ey = endY - startY;
		int dx, dy, error;

		pixNo++;
		if (ex > 0) {
			dx = 1;
		} else if (ex < 0) {
			dx = -1;
			ex = -ex;
		} else {
			dx = 0;
		}
		if (ey > 0) {
			dy = 1;
		} else if (ey < 0) {
			dy = -1;
			ey = -ey;
		} else {
			dy = 0;
		}
		pixX = startX;
		pixY = startY;
		if (ex > ey) {
			error = 2 * ey - ex;
			while (pix != pixNo) {
				if (error >= 0) {
					error -= 2 * ex;
					pixY += dy;
				}
				error += 2 * ey;
				pixX += dx;
				pix++;
				if (pixX == endX)
					return true;
			}
			return false;
		}
		error = 2 * ex - ey;
		while (pix != pixNo) {
			if (error >= 0) {
				error -= 2 * ey;
				pixX += dx;
			}
			error += 2 * ex;
			pixY += dy;
			pix++;
			if (pixY == endY)
				return true;
		}
		return false;
	}
}
